from bisect import bisect_left
from collections import deque
from html.entities import html5 as entities

from .constants import (
    EOF,
    Token,
    ascii_letters,
    ascii_upper_to_lower,
    digits,
    hexdigits,
    replacement_characters,
    space_characters,
    tag_token_types,
)
from .inputstream import HTMLInputStream

entity_keys = tuple(sorted(entities))


def has_keys_with_prefix(prefix):
    if prefix in entities:
        return True
    if (i := bisect_left(entity_keys, prefix)) == len(entities):
        return False
    return entity_keys[i].startswith(prefix)


def longest_prefix(prefix):
    if prefix in entities:
        return prefix
    for i in range(1, len(prefix) + 1):
        if prefix[:-i] in entities:
            return prefix[:-i]
    raise KeyError(prefix)


class HTMLTokenizer:
    """HTML tokenizer."""

    def __init__(self, stream, parser=None, **kwargs):
        self.stream = HTMLInputStream(stream, **kwargs)  # HTMLInputStream object
        self.parser = parser

        # Setup the initial tokenizer state
        self.state = self.data_state  # method to be invoked
        self.current_token = None  # token currently being processed

    def __iter__(self):
        """This is where the magic happens.

        We do our usually processing through the states and when we have a token
        to return we yield the token which pauses processing until the next token
        is requested.

        """
        self.token_queue = deque()
        # Start processing. When EOF is reached self.state will return False
        # instead of True and the loop will terminate.
        while self.state():
            while self.stream.errors:
                yield {
                    "type": Token.PARSE_ERROR,
                    "data": self.stream.errors.pop(0),
                }
            while self.token_queue:
                yield self.token_queue.popleft()

    def parse_error(self, _data, **datavars):
        """Add a parse error to the token queue."""
        token = {"type": Token.PARSE_ERROR, "data": _data}
        if datavars:
            token["datavars"] = datavars
        self.token_queue.append(token)

    def characters(self, _data):
        """Add a characters string to the token queue."""
        self.token_queue.append({"type": Token.CHARACTERS, "data": _data})

    def consume_number_entity(self, is_hex):
        """Return either U+FFFD or the character based on the representation.

        It also discards ";" if present. If not present self.parse_error is
        invoked.

        """
        allowed = hexdigits if is_hex else digits
        radix = 16 if is_hex else 10
        stack = []

        # Consume all the characters that are in range while making sure we
        # don't hit an EOF.
        character = self.stream.character()
        while character in allowed:
            stack.append(character)
            character = self.stream.character()

        # Convert the set of characters consumed to an int.
        integer = int("".join(stack), radix)

        # Certain characters get replaced with others
        if integer in replacement_characters:
            replacement = replacement_characters[integer]
            self.parse_error("illegal-codepoint-for-numeric-entity", integer=integer)
        elif (0xD800 <= integer <= 0xDFFF) or (integer > 0x10FFFF):
            replacement = "\uFFFD"
            self.parse_error("illegal-codepoint-for-numeric-entity", integer=integer)
        else:
            # Should speed up this check somehow (e.g. move the set to a constant).
            if ((0x0001 <= integer <= 0x0008) or
                (0x000E <= integer <= 0x001F) or
                (0x007F <= integer <= 0x009F) or
                (0xFDD0 <= integer <= 0xFDEF) or
                integer in frozenset([
                    0x000B, 0xFFFE, 0xFFFF, 0x1FFFE, 0x1FFFF, 0x2FFFE, 0x2FFFF,
                    0x3FFFE, 0x3FFFF, 0x4FFFE, 0x4FFFF, 0x5FFFE, 0x5FFFF,
                    0x6FFFE, 0x6FFFF, 0x7FFFE, 0x7FFFF, 0x8FFFE, 0x8FFFF,
                    0x9FFFE, 0x9FFFF, 0xAFFFE, 0xAFFFF, 0xBFFFE, 0xBFFFF,
                    0xCFFFE, 0xCFFFF, 0xDFFFE, 0xDFFFF, 0xEFFFE, 0xEFFFF,
                    0xFFFFE, 0xFFFFF, 0x10FFFE, 0x10FFFF])):
                self.parse_error(
                    "illegal-codepoint-for-numeric-entity", integer=integer)
            replacement = chr(integer)

        # Discard the ; if present. Otherwise, put it back on the queue and
        # invoke parse_error on parser.
        if character != ";":
            self.parse_error("numeric-entity-without-semicolon")
            self.stream.unget(character)

        return replacement

    def consume_entity(self, allowed=None, from_attribute=False):
        # Initialise to the default output for when no entity is matched.
        output = "&"

        stack = [self.stream.character()]
        unget = (
            stack[0] in (EOF, "<", "&", *space_characters) or
            (allowed is not None and allowed == stack[0]))
        if unget:
            self.stream.unget(stack[0])

        elif stack[0] == "#":
            # Read the next character to see if it's hex or decimal.
            hex = False
            stack.append(self.stream.character())
            if stack[-1] in ("x", "X"):
                hex = True
                stack.append(self.stream.character())

            # stack[-1] should be the first digit.
            if stack[-1] in (hexdigits if hex else digits):
                # At least one digit found, so consume the whole number.
                self.stream.unget(stack[-1])
                output = self.consume_number_entity(hex)
            else:
                # No digits found.
                self.parse_error("expected-numeric-entity")
                self.stream.unget(stack.pop())
                output = f"&{''.join(stack)}"

        else:
            # At this point in the process might have named entity. Entities
            # are stored in the global variable "entities". Consume characters
            # and compare to these to a substring of the entity names in the
            # list until the substring no longer matches.
            while stack[-1] is not EOF:
                if not has_keys_with_prefix("".join(stack)):
                    break
                stack.append(self.stream.character())

            # At this point we have a string that starts with some characters
            # that may match an entity
            # Try to find the longest entity the string will match to take care
            # of &noti for instance.
            try:
                entity_name = longest_prefix("".join(stack[:-1]))
            except KeyError:
                self.parse_error("expected-named-entity")
                self.stream.unget(stack.pop())
                output = f"&{''.join(stack)}"
            else:
                if entity_name[-1] != ";":
                    self.parse_error("named-entity-without-semicolon")
                entity_length = len(entity_name)
                allowed_character = (
                    stack[entity_length] in ascii_letters or
                    stack[entity_length] in digits or
                    stack[entity_length] == "=")
                if entity_name[-1] != ";" and from_attribute and allowed_character:
                    self.stream.unget(stack.pop())
                    output = f"&{''.join(stack)}"
                else:
                    self.stream.unget(stack.pop())
                    output = f"{entities[entity_name]}{''.join(stack[entity_length:])}"

        if from_attribute:
            self.current_token["data"][-1][1] += output
        else:
            type = "SPACE_CHARACTERS" if output in space_characters else "CHARACTERS"
            self.token_queue.append({"type": Token[type], "data": output})

    def process_entity_in_attribute(self, allowed):
        """Replace the need for entity_in_attribute_value_state."""
        self.consume_entity(allowed=allowed, from_attribute=True)

    def emit_current_token(self):
        """This method is a generic handler for emitting the tags.

        It also sets the state to "data" because that's what's needed after a
        token has been emitted.

        """
        token = self.current_token
        # Add token to the queue to be yielded.
        if token["type"] in tag_token_types:
            token["name"] = token["name"].translate(ascii_upper_to_lower)
            if token["type"] == Token.START_TAG:
                raw = token["data"]
                data = dict(raw)
                if len(raw) > len(data):
                    # We had some duplicated attribute, fix so first wins.
                    data.update(raw[::-1])
                token["data"] = data

            if token["type"] == Token.END_TAG:
                if token["data"]:
                    self.parse_error("attributes-in-end-tag")
                if token["selfClosing"]:
                    self.parse_error("self-closing-flag-on-end-tag")
        self.token_queue.append(token)
        self.state = self.data_state

    # Below are the various tokenizer states worked out.
    def data_state(self):
        data = self.stream.character()
        if data == "&":
            self.state = self.entity_data_state
        elif data == "<":
            self.state = self.tag_open_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\u0000")
        elif data is EOF:
            return False
        elif data in space_characters:
            # Directly after emitting a token you switch back to the "data
            # state". At that point space characters are important so they are
            # emitted separately.
            self.token_queue.append({
                "type": Token.SPACE_CHARACTERS,
                "data": data + self.stream.chars_until(space_characters, True),
            })
            # No need to update lastFourChars here, since the first space will
            # have already been appended to lastFourChars and will have broken
            # any <!-- or --> sequences.
        else:
            characters = self.stream.chars_until(("&", "<", "\u0000"))
            self.characters(data + characters)
        return True

    def entity_data_state(self):
        self.consume_entity()
        self.state = self.data_state
        return True

    def rcdata_state(self):
        data = self.stream.character()
        if data == "&":
            self.state = self.character_reference_in_rc_data_state
        elif data == "<":
            self.state = self.rcdata_less_than_sign_state
        elif data is EOF:
            # Tokenization ends.
            return False
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
        elif data in space_characters:
            # Directly after emitting a token you switch back to the "data
            # state". At that point space_characters are important so they are
            # emitted separately.
            self.token_queue.append({
                "type": Token.SPACE_CHARACTERS,
                "data": data + self.stream.chars_until(space_characters, True),
            })
            # No need to update lastFourChars here, since the first space will
            # have already been appended to lastFourChars and will have broken
            # any <!-- or --> sequences.
        else:
            chars = self.stream.chars_until(("&", "<", "\u0000"))
            self.characters(data + chars)
        return True

    def character_reference_in_rc_data_state(self):
        self.consume_entity()
        self.state = self.rcdata_state
        return True

    def rawtext_state(self):
        data = self.stream.character()
        if data == "<":
            self.state = self.rawtext_less_than_sign_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
        elif data is EOF:
            return False
        else:
            characters = self.stream.chars_until(("<", "\u0000"))
            self.characters(data + characters)
        return True

    def script_data_state(self):
        data = self.stream.character()
        if data == "<":
            self.state = self.script_data_less_than_sign_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
        elif data is EOF:
            return False
        else:
            characters = self.stream.chars_until(("<", "\u0000"))
            self.characters(data + characters)
        return True

    def plaintext_state(self):
        data = self.stream.character()
        if data is EOF:
            return False
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
        else:
            self.characters(data + self.stream.chars_until("\u0000"))
        return True

    def tag_open_state(self):
        data = self.stream.character()
        if data == "!":
            self.state = self.markup_declaration_open_state
        elif data == "/":
            self.state = self.close_tag_open_state
        elif data in ascii_letters:
            self.current_token = {
                "type": Token.START_TAG,
                "name": data,
                "data": [],
                "selfClosing": False,
                "selfClosingAcknowledged": False,
            }
            self.state = self.tag_name_state
        elif data == ">":
            # XXX In theory it could be something besides a tag name. But
            # do we really care?
            self.parse_error("expected-tag-name-but-got-right-bracket")
            self.characters("<>")
            self.state = self.data_state
        elif data == "?":
            # XXX In theory it could be something besides a tag name. But
            # do we really care?
            self.parse_error("expected-tag-name-but-got-question-mark")
            self.stream.unget(data)
            self.state = self.bogus_comment_state
        else:
            # XXX
            self.parse_error("expected-tag-name")
            self.characters("<")
            self.stream.unget(data)
            self.state = self.data_state
        return True

    def close_tag_open_state(self):
        data = self.stream.character()
        if data in ascii_letters:
            self.current_token = {
                "type": Token.END_TAG,
                "name": data,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.tag_name_state
        elif data == ">":
            self.parse_error("expected-closing-tag-but-got-right-bracket")
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("expected-closing-tag-but-got-eof")
            self.characters("</")
            self.state = self.data_state
        else:
            # XXX data can be _'_...
            self.parse_error("expected-closing-tag-but-got-char", data=data)
            self.stream.unget(data)
            self.state = self.bogus_comment_state
        return True

    def tag_name_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.state = self.before_attribute_name_state
        elif data == ">":
            self.emit_current_token()
        elif data is EOF:
            self.parse_error("eof-in-tag-name")
            self.state = self.data_state
        elif data == "/":
            self.state = self.self_closing_start_tag_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["name"] += "\uFFFD"
        else:
            self.current_token["name"] += data
            # (Don't use chars_until here, because tag names are
            # very short and it's faster to not do anything fancy.)
        return True

    def rcdata_less_than_sign_state(self):
        data = self.stream.character()
        if data == "/":
            self.temporary_buffer = ""
            self.state = self.rcdata_end_tag_open_state
        else:
            self.characters("<")
            self.stream.unget(data)
            self.state = self.rcdata_state
        return True

    def rcdata_end_tag_open_state(self):
        data = self.stream.character()
        if data in ascii_letters:
            self.temporary_buffer += data
            self.state = self.rcdata_end_tag_name_state
        else:
            self.characters("</")
            self.stream.unget(data)
            self.state = self.rcdata_state
        return True

    def rcdata_end_tag_name_state(self):
        appropriate = (
            self.current_token and
            self.current_token["name"].lower() == self.temporary_buffer.lower())
        data = self.stream.character()
        if data in space_characters and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.before_attribute_name_state
        elif data == "/" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.self_closing_start_tag_state
        elif data == ">" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.emit_current_token()
            self.state = self.data_state
        elif data in ascii_letters:
            self.temporary_buffer += data
        else:
            self.characters(f"</{self.temporary_buffer}")
            self.stream.unget(data)
            self.state = self.rcdata_state
        return True

    def rawtext_less_than_sign_state(self):
        data = self.stream.character()
        if data == "/":
            self.temporary_buffer = ""
            self.state = self.rawtext_end_tag_open_state
        else:
            self.characters("<")
            self.stream.unget(data)
            self.state = self.rawtext_state
        return True

    def rawtext_end_tag_open_state(self):
        data = self.stream.character()
        if data in ascii_letters:
            self.temporary_buffer += data
            self.state = self.rawtext_end_tag_name_state
        else:
            self.characters("</")
            self.stream.unget(data)
            self.state = self.rawtext_state
        return True

    def rawtext_end_tag_name_state(self):
        appropriate = (
            self.current_token and
            self.current_token["name"].lower() == self.temporary_buffer.lower())
        data = self.stream.character()
        if data in space_characters and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.before_attribute_name_state
        elif data == "/" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.self_closing_start_tag_state
        elif data == ">" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.emit_current_token()
            self.state = self.data_state
        elif data in ascii_letters:
            self.temporary_buffer += data
        else:
            self.characters(f"</{self.temporary_buffer}")
            self.stream.unget(data)
            self.state = self.rawtext_state
        return True

    def script_data_less_than_sign_state(self):
        data = self.stream.character()
        if data == "/":
            self.temporary_buffer = ""
            self.state = self.script_data_end_tag_open_state
        elif data == "!":
            self.characters("<!")
            self.state = self.script_data_escape_start_state
        else:
            self.characters("<")
            self.stream.unget(data)
            self.state = self.script_data_state
        return True

    def script_data_end_tag_open_state(self):
        data = self.stream.character()
        if data in ascii_letters:
            self.temporary_buffer += data
            self.state = self.script_data_end_tag_name_state
        else:
            self.characters("</")
            self.stream.unget(data)
            self.state = self.script_data_state
        return True

    def script_data_end_tag_name_state(self):
        appropriate = (
            self.current_token and
            self.current_token["name"].lower() == self.temporary_buffer.lower())
        data = self.stream.character()
        if data in space_characters and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.before_attribute_name_state
        elif data == "/" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.self_closing_start_tag_state
        elif data == ">" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.emit_current_token()
            self.state = self.data_state
        elif data in ascii_letters:
            self.temporary_buffer += data
        else:
            self.characters(f"</{self.temporary_buffer}")
            self.stream.unget(data)
            self.state = self.script_data_state
        return True

    def script_data_escape_start_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
            self.state = self.script_data_escape_start_dash_state
        else:
            self.stream.unget(data)
            self.state = self.script_data_state
        return True

    def script_data_escape_start_dash_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
            self.state = self.script_data_escaped_dash_dash_state
        else:
            self.stream.unget(data)
            self.state = self.script_data_state
        return True

    def script_data_escaped_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
            self.state = self.script_data_escaped_dash_state
        elif data == "<":
            self.state = self.script_data_escaped_less_than_sign_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
        elif data is EOF:
            self.state = self.data_state
        else:
            self.characters(data + self.stream.chars_until(("<", "-", "\u0000")))
        return True

    def script_data_escaped_dash_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
            self.state = self.script_data_escaped_dash_dash_state
        elif data == "<":
            self.state = self.script_data_escaped_less_than_sign_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
            self.state = self.script_data_escaped_state
        elif data is EOF:
            self.state = self.data_state
        else:
            self.characters(data)
            self.state = self.script_data_escaped_state
        return True

    def script_data_escaped_dash_dash_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
        elif data == "<":
            self.state = self.script_data_escaped_less_than_sign_state
        elif data == ">":
            self.characters(">")
            self.state = self.script_data_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
            self.state = self.script_data_escaped_state
        elif data is EOF:
            self.state = self.data_state
        else:
            self.characters(data)
            self.state = self.script_data_escaped_state
        return True

    def script_data_escaped_less_than_sign_state(self):
        data = self.stream.character()
        if data == "/":
            self.temporary_buffer = ""
            self.state = self.script_data_escaped_end_tag_open_state
        elif data in ascii_letters:
            self.characters(f"<{data}")
            self.temporary_buffer = data
            self.state = self.script_data_double_escape_start_state
        else:
            self.characters("<")
            self.stream.unget(data)
            self.state = self.script_data_escaped_state
        return True

    def script_data_escaped_end_tag_open_state(self):
        data = self.stream.character()
        if data in ascii_letters:
            self.temporary_buffer = data
            self.state = self.script_data_escaped_end_tag_name_state
        else:
            self.characters("</")
            self.stream.unget(data)
            self.state = self.script_data_escaped_state
        return True

    def script_data_escaped_end_tag_name_state(self):
        appropriate = (
            self.current_token and
            self.current_token["name"].lower() == self.temporary_buffer.lower())
        data = self.stream.character()
        if data in space_characters and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.before_attribute_name_state
        elif data == "/" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.state = self.self_closing_start_tag_state
        elif data == ">" and appropriate:
            self.current_token = {
                "type": Token.END_TAG,
                "name": self.temporary_buffer,
                "data": [],
                "selfClosing": False,
            }
            self.emit_current_token()
            self.state = self.data_state
        elif data in ascii_letters:
            self.temporary_buffer += data
        else:
            self.characters(f"</{self.temporary_buffer}")
            self.stream.unget(data)
            self.state = self.script_data_escaped_state
        return True

    def script_data_double_escape_start_state(self):
        data = self.stream.character()
        if data in (space_characters | frozenset(("/", ">"))):
            self.characters(data)
            if self.temporary_buffer.lower() == "script":
                self.state = self.script_data_double_escaped_state
            else:
                self.state = self.script_data_escaped_state
        elif data in ascii_letters:
            self.characters(data)
            self.temporary_buffer += data
        else:
            self.stream.unget(data)
            self.state = self.script_data_escaped_state
        return True

    def script_data_double_escaped_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
            self.state = self.script_data_double_escaped_dash_state
        elif data == "<":
            self.characters("<")
            self.state = self.script_data_double_escaped_less_than_sign_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
        elif data is EOF:
            self.parse_error("eof-in-script-in-script")
            self.state = self.data_state
        else:
            self.characters(data)
        return True

    def script_data_double_escaped_dash_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
            self.state = self.script_data_double_escaped_dash_dash_state
        elif data == "<":
            self.characters("<")
            self.state = self.script_data_double_escaped_less_than_sign_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
            self.state = self.script_data_double_escaped_state
        elif data is EOF:
            self.parse_error("eof-in-script-in-script")
            self.state = self.data_state
        else:
            self.characters(data)
            self.state = self.script_data_double_escaped_state
        return True

    def script_data_double_escaped_dash_dash_state(self):
        data = self.stream.character()
        if data == "-":
            self.characters("-")
        elif data == "<":
            self.characters("<")
            self.state = self.script_data_double_escaped_less_than_sign_state
        elif data == ">":
            self.characters(">")
            self.state = self.script_data_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.characters("\uFFFD")
            self.state = self.script_data_double_escaped_state
        elif data is EOF:
            self.parse_error("eof-in-script-in-script")
            self.state = self.data_state
        else:
            self.characters(data)
            self.state = self.script_data_double_escaped_state
        return True

    def script_data_double_escaped_less_than_sign_state(self):
        data = self.stream.character()
        if data == "/":
            self.characters("/")
            self.temporary_buffer = ""
            self.state = self.script_data_double_escape_end_state
        else:
            self.stream.unget(data)
            self.state = self.script_data_double_escaped_state
        return True

    def script_data_double_escape_end_state(self):
        data = self.stream.character()
        if data in (space_characters | frozenset(("/", ">"))):
            self.characters(data)
            if self.temporary_buffer.lower() == "script":
                self.state = self.script_data_escaped_state
            else:
                self.state = self.script_data_double_escaped_state
        elif data in ascii_letters:
            self.characters(data)
            self.temporary_buffer += data
        else:
            self.stream.unget(data)
            self.state = self.script_data_double_escaped_state
        return True

    def before_attribute_name_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.stream.chars_until(space_characters, True)
        elif data in ascii_letters:
            self.current_token["data"].append([data, ""])
            self.state = self.attribute_name_state
        elif data == ">":
            self.emit_current_token()
        elif data == "/":
            self.state = self.self_closing_start_tag_state
        elif data in ("'", '"', "=", "<"):
            self.parse_error("invalid-character-in-attribute-name")
            self.current_token["data"].append([data, ""])
            self.state = self.attribute_name_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"].append(["\uFFFD", ""])
            self.state = self.attribute_name_state
        elif data is EOF:
            self.parse_error("expected-attribute-name-but-got-eof")
            self.state = self.data_state
        else:
            self.current_token["data"].append([data, ""])
            self.state = self.attribute_name_state
        return True

    def attribute_name_state(self):
        data = self.stream.character()
        leaving_this_state = True
        emit_token = False
        if data == "=":
            self.state = self.before_attribute_value_state
        elif data in ascii_letters:
            self.current_token["data"][-1][0] += (
                data + self.stream.chars_until(ascii_letters, True))
            leaving_this_state = False
        elif data == ">":
            # XXX If we emit here the attributes are converted to a dict
            # without being checked and when the code below runs we error
            # because data is a dict not a list.
            emit_token = True
        elif data in space_characters:
            self.state = self.after_attribute_name_state
        elif data == "/":
            self.state = self.self_closing_start_tag_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"][-1][0] += "\uFFFD"
            leaving_this_state = False
        elif data in ("'", '"', "<"):
            self.parse_error("invalid-character-in-attribute-name")
            self.current_token["data"][-1][0] += data
            leaving_this_state = False
        elif data is EOF:
            self.parse_error("eof-in-attribute-name")
            self.state = self.data_state
        else:
            self.current_token["data"][-1][0] += data
            leaving_this_state = False

        if leaving_this_state:
            # Attributes are not dropped at this stage. That happens when the
            # start tag token is emitted so values can still be safely appended
            # to attributes, but we do want to report the parse error in time.
            self.current_token["data"][-1][0] = (
                self.current_token["data"][-1][0].translate(ascii_upper_to_lower))
            for name, _ in self.current_token["data"][:-1]:
                if self.current_token["data"][-1][0] == name:
                    self.parse_error("duplicate-attribute")
                    break
            # XXX Fix for above.
            if emit_token:
                self.emit_current_token()
        return True

    def after_attribute_name_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.stream.chars_until(space_characters, True)
        elif data == "=":
            self.state = self.before_attribute_value_state
        elif data == ">":
            self.emit_current_token()
        elif data in ascii_letters:
            self.current_token["data"].append([data, ""])
            self.state = self.attribute_name_state
        elif data == "/":
            self.state = self.self_closing_start_tag_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"].append(["\uFFFD", ""])
            self.state = self.attribute_name_state
        elif data in ("'", '"', "<"):
            self.parse_error("invalid-character-after-attribute-name")
            self.current_token["data"].append([data, ""])
            self.state = self.attribute_name_state
        elif data is EOF:
            self.parse_error("expected-end-of-tag-but-got-eof")
            self.state = self.data_state
        else:
            self.current_token["data"].append([data, ""])
            self.state = self.attribute_name_state
        return True

    def before_attribute_value_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.stream.chars_until(space_characters, True)
        elif data == '"':
            self.state = self.attribute_value_double_quoted_state
        elif data == "&":
            self.state = self.attribute_value_unquoted_state
            self.stream.unget(data)
        elif data == "'":
            self.state = self.attribute_value_single_quoted_state
        elif data == ">":
            self.parse_error("expected-attribute-value-but-got-right-bracket")
            self.emit_current_token()
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"][-1][1] += "\uFFFD"
            self.state = self.attribute_value_unquoted_state
        elif data in ("=", "<", "`"):
            self.parse_error("equals-in-unquoted-attribute-value")
            self.current_token["data"][-1][1] += data
            self.state = self.attribute_value_unquoted_state
        elif data is EOF:
            self.parse_error("expected-attribute-value-but-got-eof")
            self.state = self.data_state
        else:
            self.current_token["data"][-1][1] += data
            self.state = self.attribute_value_unquoted_state
        return True

    def attribute_value_double_quoted_state(self):
        data = self.stream.character()
        if data == '"':
            self.state = self.after_attribute_value_state
        elif data == "&":
            self.process_entity_in_attribute('"')
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"][-1][1] += "\uFFFD"
        elif data is EOF:
            self.parse_error("eof-in-attribute-value-double-quote")
            self.state = self.data_state
        else:
            self.current_token["data"][-1][1] += (
                data + self.stream.chars_until(('"', "&", "\u0000")))
        return True

    def attribute_value_single_quoted_state(self):
        data = self.stream.character()
        if data == "'":
            self.state = self.after_attribute_value_state
        elif data == "&":
            self.process_entity_in_attribute("'")
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"][-1][1] += "\uFFFD"
        elif data is EOF:
            self.parse_error("eof-in-attribute-value-single-quote")
            self.state = self.data_state
        else:
            self.current_token["data"][-1][1] += (
                data + self.stream.chars_until(("'", "&", "\u0000")))
        return True

    def attribute_value_unquoted_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.state = self.before_attribute_name_state
        elif data == "&":
            self.process_entity_in_attribute(">")
        elif data == ">":
            self.emit_current_token()
        elif data in ('"', "'", "=", "<", "`"):
            self.parse_error("unexpected-character-in-unquoted-attribute-value")
            self.current_token["data"][-1][1] += data
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"][-1][1] += "\uFFFD"
        elif data is EOF:
            self.parse_error("eof-in-attribute-value-no-quotes")
            self.state = self.data_state
        else:
            self.current_token["data"][-1][1] += (
                data + self.stream.chars_until(
                    frozenset(("&", ">", '"', "'", "=", "<", "`", "\u0000")) |
                    space_characters))
        return True

    def after_attribute_value_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.state = self.before_attribute_name_state
        elif data == ">":
            self.emit_current_token()
        elif data == "/":
            self.state = self.self_closing_start_tag_state
        elif data is EOF:
            self.parse_error("unexpected-eof-after-attribute-value")
            self.stream.unget(data)
            self.state = self.data_state
        else:
            self.parse_error("unexpected-character-after-attribute-value")
            self.stream.unget(data)
            self.state = self.before_attribute_name_state
        return True

    def self_closing_start_tag_state(self):
        data = self.stream.character()
        if data == ">":
            self.current_token["selfClosing"] = True
            self.emit_current_token()
        elif data is EOF:
            self.parse_error("unexpected-eof-after-solidus-in-tag")
            self.stream.unget(data)
            self.state = self.data_state
        else:
            self.parse_error("unexpected-character-after-solidus-in-tag")
            self.stream.unget(data)
            self.state = self.before_attribute_name_state
        return True

    def bogus_comment_state(self):
        # Make a new comment token and give it as value all the characters
        # until the first > or EOF (chars_until checks for EOF automatically)
        # and emit it.
        data = self.stream.chars_until(">")
        data = data.replace("\u0000", "\uFFFD")
        self.token_queue.append({"type": Token.COMMENT, "data": data})

        # Eat the character directly after the bogus comment which is either a
        # ">" or an EOF.
        self.stream.character()
        self.state = self.data_state
        return True

    def markup_declaration_open_state(self):
        stack = [self.stream.character()]
        if stack[-1] == "-":
            stack.append(self.stream.character())
            if stack[-1] == "-":
                self.current_token = {"type": Token.COMMENT, "data": ""}
                self.state = self.comment_start_state
                return True
        elif stack[-1] and stack[-1] in "dD":
            matched = True
            for expected in ("oO", "cC", "tT", "yY", "pP", "eE"):
                stack.append(self.stream.character())
                if not stack[-1] or stack[-1] not in expected:
                    matched = False
                    break
            if matched:
                self.current_token = {
                    "type": Token.DOCTYPE,
                    "name": "",
                    "publicId": None,
                    "systemId": None,
                    "correct": True,
                }
                self.state = self.doctype_state
                return True
        elif (stack[-1] == "[" and
              self.parser is not None and
              self.parser.tree.open_elements and
              self.parser.tree.open_elements[-1].namespace !=
              self.parser.tree.default_namespace):
            matched = True
            for expected in "CDATA[":
                stack.append(self.stream.character())
                if stack[-1] != expected:
                    matched = False
                    break
            if matched:
                self.state = self.cdata_section_state
                return True

        self.parse_error("expected-dashes-or-doctype")

        while stack:
            self.stream.unget(stack.pop())
        self.state = self.bogus_comment_state
        return True

    def comment_start_state(self):
        data = self.stream.character()
        if data == "-":
            self.state = self.comment_start_dash_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"] += "\uFFFD"
        elif data == ">":
            self.parse_error("incorrect-comment")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-comment")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["data"] += data
            self.state = self.comment_state
        return True

    def comment_start_dash_state(self):
        data = self.stream.character()
        if data == "-":
            self.state = self.comment_end_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"] += "-\uFFFD"
        elif data == ">":
            self.parse_error("incorrect-comment")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-comment")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["data"] += f"-{data}"
            self.state = self.comment_state
        return True

    def comment_state(self):
        data = self.stream.character()
        if data == "-":
            self.state = self.comment_end_dash_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"] += "\uFFFD"
        elif data is EOF:
            self.parse_error("eof-in-comment")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["data"] += (
                data + self.stream.chars_until(("-", "\u0000")))
        return True

    def comment_end_dash_state(self):
        data = self.stream.character()
        if data == "-":
            self.state = self.comment_end_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"] += "-\uFFFD"
            self.state = self.comment_state
        elif data is EOF:
            self.parse_error("eof-in-comment-end-dash")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["data"] += f"-{data}"
            self.state = self.comment_state
        return True

    def comment_end_state(self):
        data = self.stream.character()
        if data == ">":
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"] += "--\uFFFD"
            self.state = self.comment_state
        elif data == "!":
            self.parse_error("unexpected-bang-after-double-dash-in-comment")
            self.state = self.comment_end_bang_state
        elif data == "-":
            self.parse_error("unexpected-dash-after-double-dash-in-comment")
            self.current_token["data"] += data
        elif data is EOF:
            self.parse_error("eof-in-comment-double-dash")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            # XXX
            self.parse_error("unexpected-char-in-comment")
            self.current_token["data"] += f"--{data}"
            self.state = self.comment_state
        return True

    def comment_end_bang_state(self):
        data = self.stream.character()
        if data == ">":
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data == "-":
            self.current_token["data"] += "--!"
            self.state = self.comment_end_dash_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["data"] += "--!\uFFFD"
            self.state = self.comment_state
        elif data is EOF:
            self.parse_error("eof-in-comment-end-bang-state")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["data"] += f"--!{data}"
            self.state = self.comment_state
        return True

    def doctype_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.state = self.before_doctype_name_state
        elif data is EOF:
            self.parse_error("expected-doctype-name-but-got-eof")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.parse_error("need-space-after-doctype")
            self.stream.unget(data)
            self.state = self.before_doctype_name_state
        return True

    def before_doctype_name_state(self):
        data = self.stream.character()
        if data in space_characters:
            pass
        elif data == ">":
            self.parse_error("expected-doctype-name-but-got-right-bracket")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["name"] = "\uFFFD"
            self.state = self.doctype_name_state
        elif data is EOF:
            self.parse_error("expected-doctype-name-but-got-eof")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["name"] = data
            self.state = self.doctype_name_state
        return True

    def doctype_name_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.current_token["name"] = (
                self.current_token["name"].translate(ascii_upper_to_lower))
            self.state = self.after_doctype_name_state
        elif data == ">":
            self.current_token["name"] = (
                self.current_token["name"].translate(ascii_upper_to_lower))
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["name"] += "\uFFFD"
            self.state = self.doctype_name_state
        elif data is EOF:
            self.parse_error("eof-in-doctype-name")
            self.current_token["correct"] = False
            self.current_token["name"] = (
                self.current_token["name"].translate(ascii_upper_to_lower))
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["name"] += data
        return True

    def after_doctype_name_state(self):
        data = self.stream.character()
        if data in space_characters:
            pass
        elif data == ">":
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.current_token["correct"] = False
            self.stream.unget(data)
            self.parse_error("eof-in-doctype")
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            if data and data in "pP":
                matched = True
                for expected in ("uU", "bB", "lL", "iI", "cC"):
                    data = self.stream.character()
                    if not data or data not in expected:
                        matched = False
                        break
                if matched:
                    self.state = self.after_doctype_public_keyword_state
                    return True
            elif data and data in "sS":
                matched = True
                for expected in ("yY", "sS", "tT", "eE", "mM"):
                    data = self.stream.character()
                    if not data or data not in expected:
                        matched = False
                        break
                if matched:
                    self.state = self.after_doctype_system_keyword_state
                    return True

            # All the characters read before the current 'data' will be
            # [a-zA-Z], so they're garbage in the bogus doctype and can be
            # discarded; only the latest character might be '>' or EOF and
            # needs to be ungetted.
            self.stream.unget(data)
            self.parse_error("expected-space-or-right-bracket-in-doctype", data=data)
            self.current_token["correct"] = False
            self.state = self.bogus_doctype_state
        return True

    def after_doctype_public_keyword_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.state = self.before_doctype_public_identifier_state
        elif data in ("'", '"'):
            self.parse_error("unexpected-char-in-doctype")
            self.stream.unget(data)
            self.state = self.before_doctype_public_identifier_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.stream.unget(data)
            self.state = self.before_doctype_public_identifier_state
        return True

    def before_doctype_public_identifier_state(self):
        data = self.stream.character()
        if data in space_characters:
            pass
        elif data == '"':
            self.current_token["publicId"] = ""
            self.state = self.doctype_public_identifier_double_quoted_state
        elif data == "'":
            self.current_token["publicId"] = ""
            self.state = self.doctype_public_identifier_single_quoted_state
        elif data == ">":
            self.parse_error("unexpected-end-of-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.parse_error("unexpected-char-in-doctype")
            self.current_token["correct"] = False
            self.state = self.bogus_doctype_state
        return True

    def doctype_public_identifier_double_quoted_state(self):
        data = self.stream.character()
        if data == '"':
            self.state = self.after_doctype_public_identifier_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["publicId"] += "\uFFFD"
        elif data == ">":
            self.parse_error("unexpected-end-of-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["publicId"] += data
        return True

    def doctype_public_identifier_single_quoted_state(self):
        data = self.stream.character()
        if data == "'":
            self.state = self.after_doctype_public_identifier_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["publicId"] += "\uFFFD"
        elif data == ">":
            self.parse_error("unexpected-end-of-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["publicId"] += data
        return True

    def after_doctype_public_identifier_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.state = self.between_doctype_public_and_system_identifiers_state
        elif data == ">":
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data == '"':
            self.parse_error("unexpected-char-in-doctype")
            self.current_token["systemId"] = ""
            self.state = self.doctype_system_identifier_double_quoted_state
        elif data == "'":
            self.parse_error("unexpected-char-in-doctype")
            self.current_token["systemId"] = ""
            self.state = self.doctype_system_identifier_single_quoted_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.parse_error("unexpected-char-in-doctype")
            self.current_token["correct"] = False
            self.state = self.bogus_doctype_state
        return True

    def between_doctype_public_and_system_identifiers_state(self):
        data = self.stream.character()
        if data in space_characters:
            pass
        elif data == ">":
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data == '"':
            self.current_token["systemId"] = ""
            self.state = self.doctype_system_identifier_double_quoted_state
        elif data == "'":
            self.current_token["systemId"] = ""
            self.state = self.doctype_system_identifier_single_quoted_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.parse_error("unexpected-char-in-doctype")
            self.current_token["correct"] = False
            self.state = self.bogus_doctype_state
        return True

    def after_doctype_system_keyword_state(self):
        data = self.stream.character()
        if data in space_characters:
            self.state = self.before_doctype_system_identifier_state
        elif data in ("'", '"'):
            self.parse_error("unexpected-char-in-doctype")
            self.stream.unget(data)
            self.state = self.before_doctype_system_identifier_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.stream.unget(data)
            self.state = self.before_doctype_system_identifier_state
        return True

    def before_doctype_system_identifier_state(self):
        data = self.stream.character()
        if data in space_characters:
            pass
        elif data == '"':
            self.current_token["systemId"] = ""
            self.state = self.doctype_system_identifier_double_quoted_state
        elif data == "'":
            self.current_token["systemId"] = ""
            self.state = self.doctype_system_identifier_single_quoted_state
        elif data == ">":
            self.parse_error("unexpected-char-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.parse_error("unexpected-char-in-doctype")
            self.current_token["correct"] = False
            self.state = self.bogus_doctype_state
        return True

    def doctype_system_identifier_double_quoted_state(self):
        data = self.stream.character()
        if data == '"':
            self.state = self.after_doctype_system_identifier_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["systemId"] += "\uFFFD"
        elif data == ">":
            self.parse_error("unexpected-end-of-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["systemId"] += data
        return True

    def doctype_system_identifier_single_quoted_state(self):
        data = self.stream.character()
        if data == "'":
            self.state = self.after_doctype_system_identifier_state
        elif data == "\u0000":
            self.parse_error("invalid-codepoint")
            self.current_token["systemId"] += "\uFFFD"
        elif data == ">":
            self.parse_error("unexpected-end-of-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.current_token["systemId"] += data
        return True

    def after_doctype_system_identifier_state(self):
        data = self.stream.character()
        if data in space_characters:
            pass
        elif data == ">":
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            self.parse_error("eof-in-doctype")
            self.current_token["correct"] = False
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            self.parse_error("unexpected-char-in-doctype")
            self.state = self.bogus_doctype_state
        return True

    def bogus_doctype_state(self):
        data = self.stream.character()
        if data == ">":
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        elif data is EOF:
            # XXX EMIT
            self.stream.unget(data)
            self.token_queue.append(self.current_token)
            self.state = self.data_state
        else:
            pass
        return True

    def cdata_section_state(self):
        data = []
        while True:
            data.append(self.stream.chars_until("]"))
            data.append(self.stream.chars_until(">"))
            char = self.stream.character()
            if char is EOF:
                break
            else:
                assert char == ">"
                if data[-1][-2:] == "]]":
                    data[-1] = data[-1][:-2]
                    break
                else:
                    data.append(char)

        data = "".join(data)
        # Deal with null here rather than in the parser.
        if (null_count := data.count("\u0000")) > 0:
            for _ in range(null_count):
                self.parse_error("invalid-codepoint")
            data = data.replace("\u0000", "\uFFFD")
        if data:
            self.characters(data)
        self.state = self.data_state
        return True



"""Tree builder."""

from copy import copy
from xml.etree import ElementTree

from .constants import namespaces, scoping_elements, table_insert_mode_elements

# The scope markers are inserted when entering object elements,
# marquees, table cells, and table captions, and are used to prevent formatting
# from "leaking" into tables, object elements, and marquees.
Marker = None

_html = namespaces["html"]
_list_elements = {
    None: (frozenset(scoping_elements), False),
    "button": (frozenset(scoping_elements | {(_html, "button")}), False),
    "list": (frozenset(scoping_elements | {(_html, "ol"), (_html, "ul")}), False),
    "table": (frozenset([(_html, "html"), (_html, "table")]), False),
    "select": (frozenset([(_html, "optgroup"), (_html, "option")]), True),
}
# XXX td, th and tr are not actually needed.
_implied_end_tags = frozenset(("dd", "dt", "li", "option", "optgroup", "p", "rp", "rt"))


class ActiveFormattingElements(list):
    def append(self, node):
        """Append node to the end of the list."""
        equal_count = 0
        if node is not Marker:
            for element in self[::-1]:
                if element is Marker:
                    break
                nodes_equal = (
                    element.name_tuple == node.name_tuple and
                    element.attributes == node.attributes)
                if nodes_equal:
                    equal_count += 1
                if equal_count == 3:
                    self.remove(element)
                    break
        list.append(self, node)


class Element:
    def __init__(self, name, namespace=None):
        self.name = name
        self.namespace = namespace
        self._element = ElementTree.Element(self._get_etree_tag(name, namespace))
        if namespace is None:
            self.name_tuple = _html, self.name
        else:
            self.name_tuple = self.namespace, self.name
        self._children = []

        # The parent of the current node (or None for the document node).
        self.parent = None

    def _get_etree_tag(self, name, namespace):
        return name if namespace is None else f"{{{namespace}}}{name}"

    def _get_attributes(self):
        return self._element.attrib

    def _set_attributes(self, attributes):
        element_attributes = self._element.attrib
        element_attributes.clear()
        if attributes:
            # Calling .items _always_ allocates, and the above truthy check is
            # cheaper than the allocation on average.
            for key, value in attributes.items():
                name = f"{{{key[2]}}}{key[1]}" if isinstance(key, tuple) else key
                element_attributes[name] = value

    # A dict holding name -> value pairs for attributes of the node.
    attributes = property(_get_attributes, _set_attributes)

    def _get_children(self):
        return self._children

    def _set_children(self, value):
        del self._element[:]
        self._children = []
        for element in value:
            self.insert_child(element)

    # A list of child nodes of the current node. This must include all
    # elements but not necessarily other node types.
    children = property(_get_children, _set_children)

    def has_content(self):
        """Return True if the node has children or text, False otherwise."""
        return bool(self._element.text or len(self._element))

    def append_child(self, node):
        """Insert node as a child of the current node."""
        self._children.append(node)
        self._element.append(node._element)
        node.parent = self

    def insert_before(self, node, reference):
        """Insert node as a child of the current node, before reference.

        Raise ValueError if reference is not a child of the current node.

        """
        index = list(self._element).index(reference._element)
        self._element.insert(index, node._element)
        node.parent = self

    def remove_child(self, node):
        """Remove node from the children of the current node."""
        self._children.remove(node)
        self._element.remove(node._element)
        node.parent = None

    def insert_text(self, text, insert_before=None):
        """Insert data as text in the current node.

        Text is positioned before the start of node insert_before or to the end
        of the node's text.

        If insert_before is a node, insert the text before this node.

        """
        if not len(self._element):
            if not self._element.text:
                self._element.text = ""
            self._element.text += text
        elif insert_before is None:
            # Insert the text as the tail of the last child element
            if not self._element[-1].tail:
                self._element[-1].tail = ""
            self._element[-1].tail += text
        else:
            # Insert the text before the specified node
            children = list(self._element)
            index = children.index(insert_before._element)
            if index > 0:
                if not self._element[index - 1].tail:
                    self._element[index - 1].tail = ""
                self._element[index - 1].tail += text
            else:
                if not self._element.text:
                    self._element.text = ""
                self._element.text += text

    def clone(self):
        """Return a shallow copy of the current node.

        The node has the same name and attributes, but no parent or children.

        """
        element = type(self)(self.name, self.namespace)
        if self._element.attrib:
            element._element.attrib = copy(self._element.attrib)
        return element

    def reparent_children(self, parent):
        """Move all the children of the current node to parent.

        This is needed so that trees that don't store text as nodes move the
        text in the correct way.

        """
        if parent.children:
            parent.children[-1]._element.tail += self._element.text
        else:
            if not parent._element.text:
                parent._element.text = ""
            if self._element.text is not None:
                parent._element.text += self._element.text
        self._element.text = ""
        for child in self.children:
            parent.append_child(child)
        self.children = []


class Comment(Element):
    def __init__(self, data):
        # Use the superclass constructor to set all properties on the
        # wrapper element
        self._element = ElementTree.Comment(data)
        self.parent = None
        self._children = []


class DocumentType(Element):
    def __init__(self, name, public_id, system_id):
        Element.__init__(self, "<!DOCTYPE>")
        self._element.text = name
        self._element.set("publicId", public_id)
        self._element.set("systemId", system_id)
        self.public_id = public_id
        self.system_id = system_id


class Document(Element):
    def __init__(self):
        Element.__init__(self, "DOCUMENT_ROOT")


class DocumentFragment(Element):
    def __init__(self):
        Element.__init__(self, "DOCUMENT_FRAGMENT")


class TreeBuilder:
    """Tree builder."""

    def __init__(self, namespace_html_elements):
        """Create a TreeBuilder.

        If namespace_html_elements is True, namespace HTML elements.

        """
        if namespace_html_elements:
            self.default_namespace = "http://www.w3.org/1999/xhtml"
        else:
            self.default_namespace = None
        self.reset()

    def reset(self):
        self.open_elements = []
        self.active_formatting_elements = ActiveFormattingElements()

        self.head_element = None
        self.form_element = None

        self.insert_from_table = False

        self.document = Document()

    def element_in_scope(self, target, variant=None):
        # If we pass a node in we match that. If we pass a string
        # match any node with that name.
        exact_node = hasattr(target, "name_tuple")
        if not exact_node:
            if isinstance(target, str):
                target = (_html, target)
            assert isinstance(target, tuple)

        list_elements, invert = _list_elements[variant]

        for node in reversed(self.open_elements):
            if exact_node and node == target:
                return True
            elif not exact_node and node.name_tuple == target:
                return True
            elif (invert ^ (node.name_tuple in list_elements)):
                return False

        # We should never reach this point.
        raise ValueError  # pragma: no cover

    def reconstruct_active_formatting_elements(self):
        # Within this algorithm the order of steps described in the
        # specification is not quite the same as the order of steps in the
        # code. It should still do the same though.

        # Step 1: stop the algorithm when there's nothing to do.
        if not self.active_formatting_elements:
            return

        # Step 2 and step 3: we start with the last element. So i is -1.
        i = len(self.active_formatting_elements) - 1
        entry = self.active_formatting_elements[i]
        if entry is Marker or entry in self.open_elements:
            return

        # Step 6.
        while entry is not Marker and entry not in self.open_elements:
            if i == 0:
                # This will be reset to 0 below.
                i = -1
                break
            i -= 1
            # Step 5: let entry be one earlier in the list.
            entry = self.active_formatting_elements[i]

        while True:
            # Step 7.
            i += 1

            # Step 8.
            entry = self.active_formatting_elements[i]
            clone = entry.clone()  # mainly to get a new copy of the attributes

            # Step 9.
            element = self.insert_element({
                "type": "StartTag",
                "name": clone.name,
                "namespace": clone.namespace,
                "data": clone.attributes,
            })

            # Step 10.
            self.active_formatting_elements[i] = element

            # Step 11.
            if element == self.active_formatting_elements[-1]:
                break

    def clear_active_formatting_elements(self):
        entry = self.active_formatting_elements.pop()
        while self.active_formatting_elements and entry is not Marker:
            entry = self.active_formatting_elements.pop()

    def element_in_active_formatting_elements(self, name):
        """Find name between end of active formatting elements and last marker.

        If an element with this name exists, return it. Else return False.

        """
        for item in self.active_formatting_elements[::-1]:
            # Check for Marker first because if it's a Marker it doesn't have a
            # name attribute.
            if item is Marker:
                break
            elif item.name == name:
                return item
        return False

    def insert_root(self, token):
        element = self.create_element(token)
        self.open_elements.append(element)
        self.document.append_child(element)

    def insert_doctype(self, token):
        name = token["name"]
        public_id = token["publicId"]
        system_id = token["systemId"]

        doctype = DocumentType(name, public_id, system_id)
        self.document.append_child(doctype)

    def insert_comment(self, token, parent):
        parent.append_child(Comment(token["data"]))

    def create_element(self, token):
        """Create an element but don't insert it anywhere."""
        name = token["name"]
        namespace = token.get("namespace", self.default_namespace)
        element = Element(name, namespace)
        element.attributes = token["data"]
        return element

    def _get_insert_from_table(self):
        return self._insert_from_table

    def _set_insert_from_table(self, value):
        """Switch the function used to insert an element."""
        self._insert_from_table = value
        if value:
            self.insert_element = self.insert_element_table
        else:
            self.insert_element = self.insert_element_normal

    insert_from_table = property(_get_insert_from_table, _set_insert_from_table)

    def insert_element_normal(self, token):
        name = token["name"]
        assert isinstance(name, str), f"Element {name} not unicode"
        namespace = token.get("namespace", self.default_namespace)
        element = Element(name, namespace)
        element.attributes = token["data"]
        self.open_elements[-1].append_child(element)
        self.open_elements.append(element)
        return element

    def insert_element_table(self, token):
        """Create an element and insert it into the tree."""
        element = self.create_element(token)
        if self.open_elements[-1].name not in table_insert_mode_elements:
            return self.insert_element_normal(token)
        else:
            # We should be in the InTable mode. This means we want to do
            # special magic element rearranging.
            parent, insert_before = self.get_table_misnested_node_position()
            if insert_before is None:
                parent.append_child(element)
            else:
                parent.insert_before(element, insert_before)
            self.open_elements.append(element)
        return element

    def insert_text(self, data, parent=None):
        """Insert text data."""
        if parent is None:
            parent = self.open_elements[-1]

        in_table = (
            self.insert_from_table and
            self.open_elements[-1].name in table_insert_mode_elements)
        if in_table:
            # We should be in the InTable mode. This means we want to do
            # special magic element rearranging.
            parent, insert_before = self.get_table_misnested_node_position()
            parent.insert_text(data, insert_before)
        else:
            parent.insert_text(data)

    def get_table_misnested_node_position(self):
        """Get foster parent element and sibling (or None) to insert before."""

        # The foster parent element is the one which comes before the most
        # recently opened table element.

        # XXX - this is really inelegant.
        last_table = None
        foster_parent = None
        insert_before = None
        for element in self.open_elements[::-1]:
            if element.name == "table":
                last_table = element
                break
        if last_table:
            # XXX - we should really check that this parent is actually a
            # node here.
            if last_table.parent:
                foster_parent = last_table.parent
                insert_before = last_table
            else:
                index = self.open_elements.index(last_table) - 1
                foster_parent = self.open_elements[index]
        else:
            foster_parent = self.open_elements[0]
        return foster_parent, insert_before

    def generate_implied_end_tags(self, exclude=None):
        name = self.open_elements[-1].name
        if name in _implied_end_tags and name != exclude:
            self.open_elements.pop()
            # XXX This is not entirely what the specification says. We should
            # investigate it more closely.
            self.generate_implied_end_tags(exclude)

    def get_document(self, full_tree=False):
        """Return the final tree."""
        if full_tree:
            return self.document._element
        return self.document._element.find(
            "html" if self.default_namespace is None else
            f"{{{self.default_namespace}}}html")

    def get_fragment(self):
        """Return the final fragment."""
        fragment = DocumentFragment()
        self.open_elements[0].reparent_children(fragment)
        return fragment._element
